from django.urls import path

from . import views

urlpatterns = [

    path('login/', views.login_view, name='login'),
    path('logout/', views.logout, name='logout'),

    path('sso/test/', views.for_logged_only_view),
    path('sso/accept/', views.authorize_from_sso_view),
    path('sso/event/', views.event_acceptor_view),
]



import logging
from functools import wraps
from typing import List

from django.contrib.auth import get_user_model

from django_sso import deauthenticate_user
from datetime import datetime


def acceptor(f):
    """
    Event receiver method decorator. If a method is written to receive events, you must decorate it,
    otherwise the system will not resolve the method as an event handler.

    This is for security reasons that only handlers can be called, not functions.
    """

    @wraps(f)
    def wrapper(self, *args, **kwargs):
        started_at = datetime.now()

        try:
            f(self, *args, **kwargs)
        except Exception as e:
            logging.error(f'Django SSO Error {f.__qualname__}: {e}')

        duration_seconds = (datetime.now() - started_at).total_seconds()

        if duration_seconds >= 1:
            logging.warning(
                f'Django SSO: Execution duration of method {f.__qualname__}'
                f' too long: {duration_seconds:.3f}s'
            )

    wrapper.__is_sso_event_acceptor = True

    return wrapper


class EventAcceptor:
    """
    This class processes all events received from SSO service
    You can override it for change behavior

    Here the method name is equal to the event type. All event handlers must be decorated with @acceptor,
    or they will not be resolved as event receiver methods.
    """

    def _get_user(self, user_identy):
        """
        Get user by USERNAME_FIELD (unique field with always filled user identy)
        By default in Django - username. Name of field stored in USERNAME_FIELD
        """
        return get_user_model().objects.filter(**{
            f'{get_user_model().USERNAME_FIELD}': user_identy
        }).first()

    @acceptor
    def update_account(self, fields: dict):
        """
        Update or create user by identy

        Event fields 'is_active', 'is_staff', 'is_superuser' will cast to bool.

        Args:
            fields (dict): Array of fields for create/update user
        """
        user_model = get_user_model()
        data = {}

        for basic_field in ('is_active', 'is_staff', 'is_superuser'):
            if (
                hasattr(user_model, basic_field)
                and basic_field in fields
            ):
                data[basic_field] = bool(fields[basic_field])

        user_model.objects.update_or_create(**{
            user_model.USERNAME_FIELD: fields['user_identy'],
            'defaults': data
        })

    @acceptor
    def deauthenticate(self, user_identy):
        """
        Process deauthentication event from gateway

        Args:
            user_identy: It's user identy. E-Mail or Login...
        """
        deauthenticate_user(user_identy)

    @acceptor
    def update_fields(self, user_identities: List[str], fields: dict):
        """
        Accepts part of additional fields (sliced per one model)

        @param user_identities: Array of affected user identities
        @param fields: Updated fields, related to the user_identities
        """

    @acceptor
    def delete_user(self, user_identy: str):
        """
        Accepts information about user deletion on the SSO gateway

        @param user_identy
        """
        user = self._get_user(user_identy)

        if not user:
            return

        if hasattr(user, 'is_active') and user.is_active:
            user.is_active = False
            user.save()

    @acceptor
    def change_user_identy(self, old, new):
        """
        Renaming user when on the SSO gateway user identy (username or email or etc...) has been changed.

        @param old: old user identy
        @param new: new user identy
        """
        user = self._get_user(old)

        if user:
            setattr(user, get_user_model().USERNAME_FIELD, new)
            user.save()



import logging

from django.contrib.auth import user_logged_out
from django.dispatch import receiver

from django_sso.sso_service import request_deauthentication


@receiver(user_logged_out)
def push_deauthenticate_user_event(user, *args, **kwargs):
    try:
        request_deauthentication(user)
    except Exception as e:
        logging.critical(f"SSO deauthentication event send failed to gateway: {e}")


import importlib
import urllib.parse
import json

from django.conf import settings
from django.contrib.auth import login, get_user_model
from django.contrib.auth.decorators import login_required
from django.core.handlers.wsgi import WSGIRequest
from django.http import JsonResponse, HttpResponse
from django.shortcuts import render, redirect
from django.utils.translation import gettext_lazy as _
from django.views.decorators.csrf import csrf_exempt
import django.contrib.auth

from django_sso.exceptions import SSOException
from django_sso.sso_gateway import set_sso_authorization_request_used
from django_sso.sso_service import get_sso_authorization_request, request_sso_authorization_request, \
    request_deauthentication


def login_view(request):
    """
    The view to redirect to SSO for authentication
    """
    try:
        sso_token = request_sso_authorization_request(request)
        request.session['token'] = sso_token
    except SSOException as e:
        return render(
            request,
            template_name='django_sso/token_validation_error.html',
            context={
                'error': str(e),
                'debug': settings.DEBUG
            }
        )

    params = urllib.parse.urlencode({'sso': sso_token})

    return render(
        request,
        context={
            'redirect_url': f"{settings.SSO['ROOT']}/login/?{params}"
        },
        template_name='django_sso/redirect_to_sso.html'
    )


@login_required
def for_logged_only_view(request):
    """
    The view for testing mechanism
    """
    user_model = get_user_model()
    return HttpResponse(f'{_("You authorized as")} <b>{getattr(request.user, user_model.USERNAME_FIELD)}</b>.')


def authorize_from_sso_view(request: WSGIRequest):
    """
    Авторизует пользователя, который вернулся от SSO-сервера
    """
    if not request.session.session_key or not request.session.get('token', ''):
        return redirect('/')

    try:
        authorization_request = get_sso_authorization_request(request.session.get('token'))

        user_model = get_user_model()

        user = user_model.objects.filter(
            **{f'{user_model.USERNAME_FIELD}': authorization_request['user_identy']}
        ).first()

        if not user:
            raise SSOException(_('Local user for SSO does not exist'))

        set_sso_authorization_request_used(request.session.get('token'))

        login(request, user)

        if authorization_request['next_url']:
            return redirect(authorization_request['next_url'])
        else:
            return redirect('/')
    except SSOException as e:
        if settings.DEBUG:
            return render(request, 'django_sso/token_validation_error.html', context={
                'error': e
            })
        else:
            return redirect('/?sso_broken_token=true')


def logout(request):
    if not request.user.is_anonymous:
        try:
            django.contrib.auth.logout(request)
        except Exception as e:
            if settings.DEBUG:
                return render(request, template_name='django_sso/deauthentication_error.html', context={
                    'error': e
                })

    return redirect('/')


@csrf_exempt
def event_acceptor_view(request):
    if request.method != 'POST':
        return HttpResponse(status=405)

    try:
        data = json.loads(request.body.decode('utf8'))
    except:
        return HttpResponse(status=400)

    if (
            not data.get('token', '').strip()
            or data.pop('token') != settings.SSO['TOKEN']
    ):
        return JsonResponse({
            'error': _('Token not provided or incorrect')
        })

    if not data.get('type', '').strip():
        return JsonResponse({'error': _('Event type field not set')})

    try:
        type_name = str(data.pop('type')).strip()

        if type_name.startswith('_'):
            return JsonResponse({'error': _('Incorrect event type name')})

        module_name, class_name = settings.SSO.get(
            'EVENT_ACCEPTOR_CLASS',
            'django_sso.sso_service.backend.EventAcceptor'
        ).rsplit('.', 1)

        dispatcher_class = getattr(importlib.import_module(module_name), class_name)

        if (
            hasattr(dispatcher_class, type_name)
            and hasattr(getattr(dispatcher_class, type_name), '__is_sso_event_acceptor')
            and getattr(dispatcher_class, type_name).__is_sso_event_acceptor
        ):
            try:
                getattr(dispatcher_class(), type_name)(**data)
            except Exception as e:
                return JsonResponse({'error': str(e)})

            return JsonResponse({'ok': True})
        else:
            return JsonResponse({'error': f"{_('Event type not supported')} ({type_name})"})

    except Exception as e:
        return JsonResponse({'error': str(e)})



import sys
from inspect import isfunction

import requests
from django.conf import settings
from django.contrib.auth import get_user_model
from django.db.models import Field
from django.db.models.base import ModelBase
from django.db.models.fields.related_descriptors import (
    ForwardOneToOneDescriptor,
    ForwardManyToOneDescriptor,
    ReverseOneToOneDescriptor,
    DeferredAttribute,
)
from django.utils.translation import gettext_lazy as _

from django_sso.exceptions import SSOException


def set_sso_authorization_request_used(sso_token):
    """
    For sso_service side. Makes SSO request as used for authentication procedure (not available for next authentications)
    """
    try:
        result = requests.post(settings.SSO['ROOT'].rstrip('/') + '/sso/make_used/', {
            'token': settings.SSO['TOKEN'],
            'authentication_token': sso_token
        })

        if result.status_code != 200:
            raise Exception(f'Некорректный ответ сервера авторизации: STATUS={result.status_code}; TEXT={result.text}')
    except Exception as e:
        raise SSOException(e)


class Settings:
    """
    The Django SSO settings provider (singleton)
    """

    instance = None
    """
    Instance of SSO settings
    """

    def __new__(cls, *args, **kwargs):
        if cls.instance is None:
            cls.instance = super().__new__(cls, *args, **kwargs)

        return cls.instance

    def __init__(self):
        self.affected_models_fields = {}

        if not hasattr(settings, 'SSO'):
            return

        assert type(settings.SSO) == dict, _('SSO settings variable must be dict')
        """
        Models, used in SSO['ADDITIONAL_FIELDS'] parameer mapped to the SSO['ADDITIONAL_FIELDS'] values.
        """

        if 'ADDITIONAL_FIELDS' in settings.SSO:
            self.validate_additional_fields(settings.SSO['ADDITIONAL_FIELDS'])

    def validate_additional_fields(self, fields_list=None):
        """
        Validate SSO[ADDITIONAL_FIELDS] and fill the self.affected_models
        """
        user_model = get_user_model()
        result_field_names = []

        assert type(fields_list) in (list, set, tuple), \
            _('The SSO[ADDITIONAL_FIELDS] setting variable must be list, set or tuple')

        for field in fields_list:
            assert type(field) == str, _('The SSO[ADDITIONAL_FIELDS] members must be str')

            field_info = field.split(':')
            field_path = field_info[0].split('.')
            target = user_model

            result_field_names.append(field_info[0] if len(field_info) == 1 else field_info[1])

            for attribute in field_path:
                if not hasattr(target, attribute):
                    raise Exception(
                        f'Django SSO: Bad field "{field_info[0]}": {target.__class__.__name__} hasnt {attribute}'
                    )

                attribute_ref = getattr(target, attribute)

                if isinstance(attribute_ref, DeferredAttribute):
                    # Any model field
                    target = attribute_ref.field
                elif isinstance(attribute_ref, (ForwardManyToOneDescriptor, ForwardOneToOneDescriptor)):
                    # Any Foreigin or OneToOne relation
                    target = attribute_ref.field.related_model
                elif isinstance(attribute_ref, ReverseOneToOneDescriptor):
                    # Reverse OneToOne relation
                    target = attribute_ref.related.related_model
                elif isfunction(attribute_ref) or isinstance(attribute_ref, property):
                    # Allow methods as data sources and @property's
                    target = attribute_ref
                else:
                    # Reject any other type
                    raise Exception(f'Django SSO: Found bad target \'{attribute_ref}\' for \'{field}\' field')

            # region Getting the model of target field/prop/method.
            if isinstance(target, ModelBase):
                affected_model = target
            elif isinstance(target, Field):
                affected_model = target.model
            elif isfunction(target):
                affected_model = getattr(sys.modules[target.__module__], target.__qualname__.split('.')[0])
            elif isinstance(target, property):
                affected_model = getattr(sys.modules[target.fget.__module__], target.fget.__qualname__.split('.')[0])
            else:
                raise Exception(f'Django SSO: Cant determine the Model of target {target} for "{field_info[0]}"')

            if affected_model not in self.affected_models_fields:
                self.affected_models_fields[affected_model] = []

            self.affected_models_fields[affected_model].append(field)
            # endregion

        # region Check for unique
        assert len(result_field_names) == len(set(result_field_names)), \
            (
                    "Django SSO: SSO[ADDITIONAL_FIELDS] fields/aliases must be distinct. Found duplicates: %s"
                    % (', '.join({f for f in result_field_names if result_field_names.count(f) > 1}))
            )
        # endregion

        # region Check for reserved names
        used_reserved_fields = []

        for reserved_field in ('username', 'is_active', 'is_staff', 'is_superuser'):
            if result_field_names.count(reserved_field) > 0:
                used_reserved_fields.append(reserved_field)

        if len(used_reserved_fields) > 0:
            raise Exception(
                'Django SSO: SSO[ADDITIONAL_FIELDS] uses reserved fields: %s'
                % ', '.join(used_reserved_fields)
            )
        # endregion



import importlib

from django.apps import AppConfig
from django.conf import settings
from django.utils.translation import gettext_lazy as _

from django.core.exceptions import ImproperlyConfigured
from .backend import EventAcceptor


class ServiceConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'django_sso.sso_service'

    def ready(self):
        if not hasattr(settings, 'SSO'):
            raise ImproperlyConfigured(f'Django SSO: Client side requires SSO variable in settings.py')

        if not hasattr(settings, 'SSO') or not isinstance(settings.SSO, dict):
            raise ImproperlyConfigured(
                f'Django SSO: In settings.py SSO variable must be dict. (Now are "{type(settings.SSO).__name__}")'
            )

        if not hasattr(settings, 'LOGIN_URL'):
            raise ImproperlyConfigured('Django SSO: Requires LOGIN_URL setting')

        for variable in ('TOKEN', 'ROOT'):
            if variable not in settings.SSO:
                raise ImproperlyConfigured(f"SSO[{variable}] {_('settings variable not set')}")

        sso_event_acceptor_class = settings.SSO.get('EVENT_ACCEPTOR_CLASS', '').strip()

        if sso_event_acceptor_class:
            if type(sso_event_acceptor_class) != str:
                raise ImproperlyConfigured(f"SSO[EVENT_ACCEPTOR_CLASS] {_('must be string')}")

            try:
                [module_name, class_name] = sso_event_acceptor_class.rsplit('.', 1)
                module = importlib.import_module(module_name)
                class_ref = getattr(module, class_name, None)

                if not class_ref:
                    raise ImproperlyConfigured(_(
                        f'In SSO[EVENT_ACCEPTOR_CLASS] declared module has no class named {class_name}'
                    ))

                if not issubclass(class_ref, EventAcceptor):
                    raise ImproperlyConfigured(
                        f'{settings.SSO["EVENT_ACCEPTOR_CLASS"]} {_("is not inherits")} '
                        f'django_sso.sso_service.backend.EventAcceptor'
                    )
            except ImproperlyConfigured as e:
                raise e
            except Exception as e:
                raise ImproperlyConfigured(_(
                    'Can\'t import SSO event acceptor class from SSO[EVENT_ACCEPTOR_CLASS] variable'
                ))

        from . import signals



from django.urls import path
from django.views.generic import TemplateView
from django.conf import settings

from . import views


urlpatterns = [
	path('login/', views.LoginView.as_view(), name="login"),
	path('logout/', views.LogoutView.as_view(), name="logout"),

	path('sso/obtain/', views.ObtainView.as_view(), name="obtain"),
	path('sso/get/', views.GetAuthenticationRequestView.as_view(), name="get"),
	path('sso/make_used/', views.MakeUsedView.as_view(), name="make_used"),
	path('sso/deauthenticate/', views.DeauthenticateView.as_view(), name="deauthenticate_view"),
	path('__welcome/', TemplateView.as_view(template_name='django_sso/welcome.html'), name="welcome"),
]

if settings.DEBUG:
	urlpatterns.append(
		path('sso/debug/update_event/', views.DebugUpdateEvent.as_view(), name="debug_update_event"),
	)



from typing import List, Union, Set

from django.apps import AppConfig
from django.contrib.auth import get_user_model
from django.db.models import QuerySet
from django.db.models.base import ModelBase
from django.db.models.fields.related_descriptors import ForwardOneToOneDescriptor, ReverseOneToOneDescriptor, \
    ForwardManyToOneDescriptor, ReverseManyToOneDescriptor
from django.db.models.signals import post_delete, pre_save, post_save
from django.dispatch import receiver
from django.utils.translation import gettext_lazy as _

from . import Settings


class DjangoSsoConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'django_sso.sso_gateway'
    verbose_name = _('Single Sign-On')

    def ready(self):
        self.sso_settings = Settings()
        self.service_model = self.get_model('Service')

        # region Bind signals for models provided in SSO['ADDITIONAL_FIELDS']
        affected_models = self.sso_settings.affected_models_fields.keys()

        model = get_user_model()

        @receiver(pre_save, sender=model)
        def pre_save_proxy(*args, **kwargs):
            self.on_pre_save_user_model(*args, **kwargs)

        @receiver(post_save, sender=model)
        def post_save_proxy(*args, **kwargs):
            self.on_post_save_user_model(*args, **kwargs)

        @receiver(post_delete, sender=model)
        def post_delete_proxy(*args, **kwargs):
            self.on_post_delete_user(*args, **kwargs)

        if len(affected_models) > 0:
            for model in affected_models:
                if model != get_user_model():
                    @receiver(pre_save, sender=model)
                    def pre_save_proxy(*args, **kwargs):
                        self.on_pre_save_related_model(*args, **kwargs)

                    @receiver(post_delete, sender=model)
                    def post_delete_proxy(*args, **kwargs):
                        self.on_post_delete_related_model(*args, **kwargs)
        # endregion

    def on_pre_save_related_model(self, sender: ModelBase, instance, *args, **kwargs):
        """
        When direct related object saved - calculate difference in relations and send to old and to new users new data
        """
        # get direct One2One|FK (from old|new model)  ++++  reverse FK relations of MODEL

        old_direct_related_users_identities = set()
        old_object_version = instance._meta.model.objects.filter(id=instance.id).first()

        if old_object_version:
            old_direct_related_users_identities = self.get_all_related_users(
                instance=old_object_version,
                include_direct_relations=True,
            )

        new_direct_related_users_identities = self.get_all_related_users(
            instance=instance,
            include_direct_relations=True,
        )

        reverse_related_users_identities = self.get_all_related_users(
            instance=instance,
            include_reversed_relations=True
        )

        deleted_from_user_identities = old_direct_related_users_identities - new_direct_related_users_identities

        if len(deleted_from_user_identities) > 0:
            """
            Send null's to all old related users
            """
            self.service_model.cast_event_to_all_services(
                event_name='update_fields',
                to_users=self._get_users_queryset_by_identities(deleted_from_user_identities),
                instance=None
            )

        actual_related_users = new_direct_related_users_identities | reverse_related_users_identities

        if len(actual_related_users):
            """
            Send updation to all current related users to modified/created object
            """
            self.service_model.cast_event_to_all_services(
                event_name='update_fields',
                to_users=self._get_users_queryset_by_identities(actual_related_users),
                instance=instance
            )

    def on_post_delete_related_model(self, sender: ModelBase, instance, *args, **kwargs):
        """
        Emit event to all related users before model deletion: send nulls.
        """
        related_users = self.get_all_related_users(instance)

        self.service_model.cast_event_to_all_services('update_fields', to_users=related_users, instance=None)

    def get_all_related_users(
            self,
            instance: ModelBase,
            include_direct_relations=False,
            include_reversed_relations=False
    ) -> Set[str]:
        """
        Find and return all users wich DIRECT related to provided model

        If called from pre_save: for oneToOne and foreigin key will return users from instance, but not from database,
        if you want to get old objects on pre_save - pass the instance obtained from DB instead of the instance from
        signal

        @param include_direct_relations:
        @param include_reversed_relations:
        @type instance: ModelBase Model wich direct related to UserModel via any relation
        @return Set[str]: Unique affected user identities
        """
        assert instance._meta.model != get_user_model(),\
            "Django SSO: get_all_related_users() accepts only direct related models"

        related_users_identities = set()

        for model_attribute_name in dir(instance._meta.model):
            if model_attribute_name[0] == '_':
                continue

            model_attribute = getattr(instance._meta.model, model_attribute_name)

            if include_direct_relations:
                if (
                    (
                        isinstance(model_attribute, (ForwardOneToOneDescriptor, ForwardManyToOneDescriptor))
                        and model_attribute.field.related_model == get_user_model()
                    ) or (
                        isinstance(model_attribute, ReverseOneToOneDescriptor)
                        and model_attribute.related.related_model == get_user_model()
                    )
                ):
                    if hasattr(instance, model_attribute_name):
                        object_attribute_value = getattr(instance, model_attribute_name)

                        if object_attribute_value:
                            related_users_identities.add(getattr(object_attribute_value, get_user_model().USERNAME_FIELD))

            if include_reversed_relations:
                if isinstance(model_attribute, ReverseManyToOneDescriptor):
                    if model_attribute.field.model == get_user_model():
                        user_identities = getattr(instance, model_attribute_name).all().values_list(
                            get_user_model().USERNAME_FIELD,
                            flat=True
                        )

                        for user_identy in user_identities:
                            related_users_identities.add(user_identy)

        return related_users_identities

    def _get_users_queryset_by_identities(self, user_identities: Union[List[str], Set[str]]) -> QuerySet:
        return get_user_model().objects.filter(**{
            f'{get_user_model().USERNAME_FIELD}__in': user_identities
        }).all()

    def on_post_delete_user(self, sender: ModelBase, instance: ModelBase, *args, **kwargs):
        self.service_model.cast_event_to_all_services('delete_user', user=instance)

    def on_post_save_user_model(self, sender: ModelBase, instance: ModelBase, *args, **kwargs):
        self.service_model.cast_event_to_all_services('update_account', user=instance)

    def on_pre_save_user_model(self, sender: ModelBase, instance: ModelBase, *args, **kwargs):
        """
        If user identy has changed - emit event about it
        """
        if instance.pk:
            old_identy = get_user_model().objects.filter(id=instance.pk).first()

            if old_identy:
                old_identy = getattr(old_identy, instance.USERNAME_FIELD)

            new_identy = getattr(instance, instance.USERNAME_FIELD)

            if old_identy != new_identy:
                self.service_model.cast_event_to_all_services('change_user_identy', old=old_identy, new=new_identy)



import importlib

from django.conf import settings
import requests
from django.contrib.auth import get_user_model

from django_sso.exceptions import SSOException


def request_sso_authorization_request(request) -> str:
    """
    Запрашивает токен авторизации на SSO-шлюзе и возвращает его как результат.
    Необходим для дальнейшей авторизации пользователя на шлюзе авторизации
    """
    try:
        result = requests.post(settings.SSO['ROOT'].rstrip('/') + '/sso/obtain/', {
            "token": settings.SSO['TOKEN'],
            "next_url": request.GET.get('next', '/'),
        })

        if result.status_code != 200:
            raise Exception(f'Некорректный ответ сервера авторизации: STATUS={result.status_code}; TEXT={result.text}')

        result = result.json()
    except Exception as e:
        raise SSOException(e)

    if 'token' in result:
        return result['token']
    else:
        raise SSOException(result['error'])


def get_sso_authorization_request(sso_token: str) -> dict:
    """
    Get SSO token information from server to check authorization
    """
    try:
        result = requests.post(settings.SSO['ROOT'].rstrip('/') + '/sso/get/', {
            'token': settings.SSO['TOKEN'],
            'authentication_token': sso_token
        })

        if result.status_code != 200:
            raise SSOException(f'Некорректный ответ сервера авторизации: STATUS={result.status_code}; TEXT={result.text}')

        result = result.json()

        if 'error' in result:
            raise SSOException(result['error'])
    except Exception as e:
        raise SSOException(e)

    return result


def request_deauthentication(user):
    """
    Call SSO sso_gateway to deauthorize user everywhere
    """
    user_model = get_user_model()

    try:
        result = requests.post(settings.SSO['ROOT'].rstrip('/') + '/sso/deauthenticate/', {
            'token': settings.SSO['TOKEN'],
            'user_identy': getattr(user, user_model.USERNAME_FIELD)
        })

        if result.status_code != 200:
            raise SSOException(
                f'Некорректный ответ сервера авторизации: STATUS={result.status_code}; TEXT={result.text}'
            )

        result = result.json()

        if 'error' in result:
            raise SSOException(result['error'])
    except Exception as e:
        raise SSOException(e)



# Generated by Django 4.1.2 on 2023-02-02 20:19

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('sso_gateway', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='authenticationrequest',
            name='used',
            field=models.BooleanField(default=False, verbose_name='Are used in external sso service'),
        ),
    ]



from setuptools import setup, find_packages

setup(
    name='django_sso',
    version='3.0.1',  # SemVer
    author="DAVIDhaker",
    author_email='me@davidhaker.ru',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    package_data={
        'django_sso': [
            '**.html'
        ]
    },
    description='Django Single Sign-On implementation',
    long_description_content_type='text/markdown',
    long_description=open('README.md', 'r', encoding='utf-8').read(),
    include_package_data=True,
    url='https://github.com/davidhaker/django-sso',
    url_download='https://pypi.org/project/django-sso/',
    keywords='Django SSO Single Sign-On',
    install_requires=[
        'django',
        'requests',
    ],
    python_requires='>=3',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Framework :: Django :: 2',
        'Framework :: Django :: 2.0',
        'Framework :: Django :: 2.1',
        'Framework :: Django :: 2.2',
        'Framework :: Django :: 3',
        'Framework :: Django :: 3.0',
        'Framework :: Django :: 3.1',
        'Framework :: Django :: 3.2',
        'Framework :: Django :: 4',
        'Framework :: Django :: 4.0',
        'Framework :: Django :: 4.1',
        'Natural Language :: English',
        'Natural Language :: Russian',
        'Programming Language :: Python :: 3',
        'Topic :: Internet :: WWW/HTTP :: Session'
    ],
)



from django import forms
from django.utils.translation import gettext_lazy as _


class LoginForm(forms.Form):
    username = forms.CharField(min_length=1, max_length=128, label=_("Login/Email"))
    password = forms.CharField(min_length=1, widget=forms.PasswordInput, label=_("Password"))



import logging
from inspect import ismethod, isfunction
from typing import Union

import requests
from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import User
from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from django.db.models import QuerySet
from django.db.models.base import ModelBase
from django.utils.crypto import get_random_string
from django.utils.translation import gettext_lazy as _

from django_sso.exceptions import SSOException
from django_sso.sso_gateway import Settings

user_model = get_user_model()


def service_token_generator():
    return get_random_string(Service.token.field.max_length)


class Service(models.Model):
    name = models.CharField(max_length=128, verbose_name=_('Name'))
    base_url = models.URLField(verbose_name=_('Base url'))
    enabled = models.BooleanField(default=False, verbose_name=_('Enabled'))
    token = models.CharField(max_length=128, verbose_name=_('Token'), unique=True, default=service_token_generator)

    def __str__(self):
        return self.base_url

    def _send_event(self, event_type, data):
        text = None
        fail = False

        if hasattr(settings, 'SSO_SUBORDINATE_COMMUNICATION_TIMEOUT'):
            timeout = settings.SSO_SUBORDINATE_COMMUNICATION_TIMEOUT

            assert type(settings.SSO_SUBORDINATE_COMMUNICATION_TIMEOUT) in (int, float)
        else:
            timeout = 0.1  # 100ms

        try:
            result = requests.post(
                f'{self.base_url}/sso/event/',
                json={
                    'type': event_type,
                    'token': self.token,
                    **data
                },
                headers={
                    "Content-Type": "application/json"
                },
                timeout=timeout
            )
        except Exception as e:
            logging.error(f"Django SSO: {_('Failed to communicate with subordinated service')} {self.base_url}: {e}")

            return

        try:
            assert result.status_code == 200, f"{result.text}"
            data = result.json()

            ok = not not (data['ok'] if 'ok' in data else False)

            if ok:
                return ok
            elif 'error' in data:
                raise Exception(f"Django SSO: {_('Error raised on subordinate service')}: {data['error']}")
            else:
                raise Exception(result.text)
        except Exception as e:
            logging.error(f'{_("Incorrect response from subordinated service")}: STATUS={result.status_code}; TEXT={e}')

            return

    @staticmethod
    def build_update_user_event(user):
        """
        Build event for accounts update on subbordinated services

        Args:
            user: AbstractBaseUser based classes are allowed
        """
        event = {
            'fields': {}
        }

        for field in ('is_active', 'is_staff', 'is_superuser'):
            if hasattr(user_model, field):
                event['fields'][field] = bool(getattr(user, field))

        event['fields']["user_identy"] = getattr(user, user_model.USERNAME_FIELD)

        if hasattr(settings, 'SSO') and 'ADDITIONAL_FIELDS' in settings.SSO:
            for additional_field in settings.SSO['ADDITIONAL_FIELDS']:
                field_info = additional_field.split(':')
                alias = field_info[1] if len(field_info) == 2 else None

                result = user

                try:
                    for prop in field_info[0].split('.'):
                        try:
                            value = getattr(result, prop)
                        except ObjectDoesNotExist:
                            value = None
                            break

                        if value != None:
                            result = value
                        else:
                            result = None
                            break

                    if ismethod(result):
                        result = result()
                    elif isinstance(result, models.Model):
                        if hasattr(result, 'to_sso_representation'):
                            result = result.to_sso_representation()
                        else:
                            result = str(result)
                except Exception as e:
                    logging.warning('Django SSO: failed to read value for field %s: %s' % (field_info[0], e))
                    result = None

                event['fields'][alias if alias else additional_field] = result

        return event

    @staticmethod
    def build_update_fields_event(user_identities: Union[set, QuerySet], instance: ModelBase):
        """
        Build event for fields update on subbordinated services. Fields of related model.

        Args:
            user_identities: QuerySet with users
            instance: An updated related model
        """

        sso_settings = Settings()

        event = {
            "fields": {},
            "user_identities": (
                [*user_identities.values_list(get_user_model().USERNAME_FIELD, flat=True)]
                if isinstance(user_identities, QuerySet)
                else user_identities
            )
        }

        for field_info in sso_settings.affected_models_fields[instance.__class__]:
            field_info = field_info.split(':')
            field_alias = field_info[1] if len(field_info) == 2 else None
            field_path = field_info[0].split('.')
            field_name = field_alias if field_alias else field_info[0]

            if instance:
                if len(field_path) == 1:
                    if hasattr(instance, 'to_sso_representation'):
                        value = instance.to_sso_representation()
                    else:
                        value = str(instance)
                elif len(field_path) == 2:
                    model_attr = getattr(instance, field_path[1])

                    if model_attr is None:
                        value = None
                    elif ismethod(model_attr):
                        value = model_attr()
                    elif isinstance(model_attr, property):
                        value = model_attr
                    else:
                        value = model_attr
                else:
                    logging.error('Django SSO: Unhandled exception. Contact developer with information about it.')

                if type(value) not in (str, bool, float, int):
                    logging.error(
                        f"Django SSO: For additional field '{field_info}' provided unsupported type {type(value)}"
                    )
                    value = None

                event['fields'][field_name] = value
            else:
                event['fields'][field_name] = None

        return event

    def deauthenticate(self, user: Union[str, ModelBase]):
        """
        Send deauthentication event to subordinate service, if that active

        Args:
            user: User model object or user identy - username field value
        """
        if not self.enabled:
            return True

        return self._send_event('deauthenticate', {
            'user_identy': user if type(user) == str else getattr(user, user_model.USERNAME_FIELD)
        })

    def delete_user(self, user: Union[ModelBase, str]):
        """
        Casts user deletion event

        @param user: User identy string or UserModel instance
        """
        return self._send_event('delete_user', {
            'user_identy': user if isinstance(user, str) else getattr(user, user_model.USERNAME_FIELD)
        })

    def change_user_identy(self, old, new):
        """
        Emit event for changing user identy.

        In cases, when you change login|email|etc...

        @param old: Old user identy
        @param new: New user identy
        """
        return self._send_event('change_user_identy', {
            'old': old,
            'new': new
        })

    def update_account(self, user) -> bool:
        """
        Send account information to subordinated service, if subordinated service is active
        """
        if not self.enabled:
            return True

        return self._send_event(
            event_type='update_account',
            data=self.build_update_user_event(user)
        )

    def update_fields(self, to_users: Union[QuerySet, set], instance: ModelBase = None) -> bool:
        """
        Send event with updated fields of the related model, if subordinated service is active
        """
        if not self.enabled or not len(to_users):
            return True

        return self._send_event(
            event_type='update_fields',
            data=self.build_update_fields_event(to_users, instance)
        )

    @staticmethod
    def cast_event_to_all_services(event_name: str, **kwargs):
        assert len(event_name) and not event_name.startswith('_'), f"Bad event name {event_name}"

        if not hasattr(Service, event_name) or not isfunction(getattr(Service, event_name)):
            raise Exception(f'Django SSO: {Service.__class__.__name__} has no method {event_name}')

        for service in Service.objects.filter(enabled=True):
            getattr(service, event_name)(**kwargs)

    class Meta:
        verbose_name = _('Subordinated service')
        verbose_name_plural = _('Subordinated services')


def auth_token_generator():
    return get_random_string(AuthenticationRequest.token.field.max_length)


class AuthenticationRequest(models.Model):
    service: Service = models.ForeignKey('Service', on_delete=models.CASCADE, verbose_name=_('Service'))
    created_at = models.DateTimeField(auto_now_add=True, verbose_name=_('Created at'))
    token = models.CharField(max_length=128, verbose_name=_('Token'), default=auth_token_generator, unique=True)
    user_identy = models.CharField(max_length=128, verbose_name=_('User identy'), help_text=_('E-Mail, login, etc.'))
    next_url = models.CharField(max_length=512, verbose_name=_('Next url'), help_text=_('To go after success auth'))
    authenticated = models.BooleanField(default=False, verbose_name=_('Request has been activated'))
    used = models.BooleanField(default=False, verbose_name=_('Are used in external sso service'))

    class Meta:
        verbose_name = _('Authentication request')
        verbose_name_plural = _('Authentication requests')

    def activate(self, user: User):
        """
        1) Activate authentication request
        2) Send base information about user to subordinated service
        """
        self.user_identy = getattr(user, user_model.USERNAME_FIELD)
        self.authenticated = True
        self.save()

        try:
            return self.service.update_account(user)
        except Exception as e:
            raise SSOException(str(e))

    def __str__(self):
        return f'{_("Authenticate")} {self.user_identy} {_("on")} {self.service} {_("then go to")} {self.next_url}'



import logging

from django.contrib.auth import user_logged_out, user_logged_in
from django.dispatch import receiver

from .models import Service


@receiver(user_logged_in)
def push_update_user_event(user=None, *args, **kwargs):
    if user:
        for service in Service.objects.filter(enabled=True):
            try:
                service.update_account(user)
            except Exception as e:
                logging.critical(f"Django SSO event dispatching error: {e}")


@receiver(user_logged_out)
def push_deauthenticate_user_event(user=None, *args, **kwargs):
    if user:
        for service in Service.objects.filter(enabled=True):
            try:
                service.deauthenticate(user)
            except Exception as e:
                logging.critical(f"Django SSO event dispatching error: {e}")
